# Generated by Django 2.2.20 on 2021-04-27 20:14

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('python', '0005_pythonpackagecontent_sha256'),
    ]

    operations = [
        migrations.AddField(
            model_name='pythonrepository',
            name='autopublish',
            field=models.BooleanField(default=False),
        ),
    ]
