# Generated by Django 2.2.23 on 2021-06-04 15:50

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('python', '0006_pythonrepository_autopublish'),
    ]

    operations = [
        migrations.AddField(
            model_name='pythonpackagecontent',
            name='description_content_type',
            field=models.TextField(default=''),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='pythonpackagecontent',
            name='project_urls',
            field=django.contrib.postgres.fields.jsonb.JSONField(default=dict),
        ),
    ]
