# Generated by Django 2.2.19 on 2021-03-26 13:31

import django.contrib.postgres.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('python', '0002_pythonpackagecontent_python_version'),
    ]

    operations = [
        migrations.AddField(
            model_name='pythonremote',
            name='exclude_platforms',
            field=django.contrib.postgres.fields.ArrayField(base_field=models.CharField(blank=True, max_length=10), choices=[('windows', 'windows'), ('macos', 'macos'), ('freebsd', 'freebsd'), ('linux', 'linux')], default=list, size=None),
        ),
        migrations.AddField(
            model_name='pythonremote',
            name='keep_latest_packages',
            field=models.IntegerField(default=0),
        ),
        migrations.AddField(
            model_name='pythonremote',
            name='package_types',
            field=django.contrib.postgres.fields.ArrayField(base_field=models.CharField(blank=True, max_length=15), choices=[('bdist_dmg', 'bdist_dmg'), ('bdist_dumb', 'bdist_dumb'), ('bdist_egg', 'bdist_egg'), ('bdist_msi', 'bdist_msi'), ('bdist_rpm', 'bdist_rpm'), ('bdist_wheel', 'bdist_wheel'), ('bdist_wininst', 'bdist_wininst'), ('sdist', 'sdist')], default=list, size=None),
        ),
    ]
